@@ -47,18 +47,17 @@
 #include "../strdup.h"
 #include "../strerror.h"
 #include "../select.h" /* for the socket readiness */
-#include "../curlx/fopen.h"
 #include "../curlx/inet_pton.h" /* for IP addr SNI check */
 #include "../curlx/multibyte.h"
 #include "../curlx/warnless.h"
 #include "x509asn1.h"
+#include "../curl_printf.h"
 #include "../multiif.h"
 #include "../system_win32.h"
 #include "../curlx/version_win32.h"
 #include "../rand.h"
 #include "../curlx/strparse.h"
 #include "../progress.h"
-#include "../curl_sha256.h"
 
 /* The last #include file should be: */
 #include "../curl_memory.h"
@@ -110,8 +109,8 @@
 #define CERT_THUMBPRINT_DATA_LEN 20
 
 /* Uncomment to force verbose output
- * #define infof(x, y, ...) curl_mprintf(y, __VA_ARGS__)
- * #define failf(x, y, ...) curl_mprintf(y, __VA_ARGS__)
+ * #define infof(x, y, ...) printf(y, __VA_ARGS__)
+ * #define failf(x, y, ...) printf(y, __VA_ARGS__)
  */
 
 /* Offered when targeting Vista (XP SP2+) */
@@ -564,7 +563,7 @@ schannel_acquire_credential_handle(struct Curl_cfilter *cf,
                                  &cert_store_path, &cert_thumbprint_str);
 
       if(result && (data->set.ssl.primary.clientcert[0]!='\0'))
-        fInCert = curlx_fopen(data->set.ssl.primary.clientcert, "rb");
+        fInCert = fopen(data->set.ssl.primary.clientcert, "rb");
 
       if(result && !fInCert) {
         failf(data, "schannel: Failed to get certificate location"
@@ -581,8 +580,6 @@ schannel_acquire_credential_handle(struct Curl_cfilter *cf,
             " for %s",
             blob ? "(memory blob)" : data->set.ssl.primary.clientcert);
       curlx_unicodefree(cert_path);
-      if(fInCert)
-        curlx_fclose(fInCert);
       return CURLE_SSL_CERTPROBLEM;
     }
 
@@ -614,7 +611,7 @@ schannel_acquire_credential_handle(struct Curl_cfilter *cf,
         if((!certdata) ||
            ((int) fread(certdata, certsize, 1, fInCert) != 1))
           continue_reading = FALSE;
-        curlx_fclose(fInCert);
+        fclose(fInCert);
         if(!continue_reading) {
           failf(data, "schannel: Failed to read cert file %s",
                 data->set.ssl.primary.clientcert);
@@ -661,7 +658,7 @@ schannel_acquire_credential_handle(struct Curl_cfilter *cf,
                 cert_showfilename_error);
         else
           failf(data, "schannel: Failed to import cert file %s, "
-                "last error is 0x%08lx",
+                "last error is 0x%lx",
                 cert_showfilename_error, errorcode);
         return CURLE_SSL_CERTPROBLEM;
       }
@@ -680,7 +677,7 @@ schannel_acquire_credential_handle(struct Curl_cfilter *cf,
 
       if(!client_certs[0]) {
         failf(data, "schannel: Failed to get certificate from file %s"
-              ", last error is 0x%08lx",
+              ", last error is 0x%lx",
               cert_showfilename_error, GetLastError());
         CertCloseStore(cert_store, 0);
         return CURLE_SSL_CERTPROBLEM;
@@ -702,7 +699,7 @@ schannel_acquire_credential_handle(struct Curl_cfilter *cf,
         char *path_utf8 =
           curlx_convert_tchar_to_UTF8(cert_store_path);
         failf(data, "schannel: Failed to open cert store %lx %s, "
-              "last error is 0x%08lx",
+              "last error is 0x%lx",
               cert_store_name,
               (path_utf8 ? path_utf8 : "(unknown)"),
               GetLastError());
@@ -821,8 +818,6 @@ schannel_acquire_credential_handle(struct Curl_cfilter *cf,
       result = set_ssl_ciphers(&schannel_cred, ciphers, algIds);
       if(result) {
         failf(data, "schannel: Failed setting algorithm cipher list");
-        if(client_certs[0])
-          CertFreeCertificateContext(client_certs[0]);
         return result;
       }
     }
@@ -850,6 +845,7 @@ schannel_acquire_credential_handle(struct Curl_cfilter *cf,
     char buffer[STRERROR_LEN];
     failf(data, "schannel: AcquireCredentialsHandle failed: %s",
           Curl_sspi_strerror(sspi_status, buffer, sizeof(buffer)));
+    Curl_safefree(backend->cred);
     switch(sspi_status) {
     case SEC_E_INSUFFICIENT_MEMORY:
       return CURLE_OUT_OF_MEMORY;
@@ -929,7 +925,7 @@ schannel_connect_step1(struct Curl_cfilter *cf, struct Curl_easy *data)
   backend->cred = NULL;
 
   /* check for an existing reusable credential handle */
-  if(Curl_ssl_scache_use(cf, data)) {
+  if(ssl_config->primary.cache_session) {
     struct Curl_schannel_cred *old_cred;
     Curl_ssl_scache_lock(data);
     old_cred = Curl_ssl_scache_get_obj(cf, data, connssl->peer.scache_key);
@@ -1115,41 +1111,6 @@ schannel_connect_step1(struct Curl_cfilter *cf, struct Curl_easy *data)
   return CURLE_OK;
 }
 
-static CURLcode schannel_error(struct Curl_easy *data,
-                               SECURITY_STATUS sspi_status)
-{
-  char buffer[STRERROR_LEN];
-  switch(sspi_status) {
-  case SEC_E_INSUFFICIENT_MEMORY:
-    failf(data, "schannel: next InitializeSecurityContext failed: %s",
-          Curl_sspi_strerror(sspi_status, buffer, sizeof(buffer)));
-    return CURLE_OUT_OF_MEMORY;
-  case SEC_E_WRONG_PRINCIPAL:
-    failf(data, "schannel: SNI or certificate check failed: %s",
-          Curl_sspi_strerror(sspi_status, buffer, sizeof(buffer)));
-    return CURLE_PEER_FAILED_VERIFICATION;
-  case SEC_E_UNTRUSTED_ROOT:
-    failf(data, "schannel: %s",
-          Curl_sspi_strerror(sspi_status, buffer, sizeof(buffer)));
-    return CURLE_PEER_FAILED_VERIFICATION;
-#if 0
-  case SEC_E_INVALID_HANDLE:
-  case SEC_E_INVALID_TOKEN:
-  case SEC_E_LOGON_DENIED:
-  case SEC_E_TARGET_UNKNOWN:
-  case SEC_E_NO_AUTHENTICATING_AUTHORITY:
-  case SEC_E_INTERNAL_ERROR:
-  case SEC_E_NO_CREDENTIALS:
-  case SEC_E_UNSUPPORTED_FUNCTION:
-  case SEC_E_APPLICATION_PROTOCOL_MISMATCH:
-#endif
-  default:
-    failf(data, "schannel: next InitializeSecurityContext failed: %s",
-          Curl_sspi_strerror(sspi_status, buffer, sizeof(buffer)));
-    return CURLE_SSL_CONNECT_ERROR;
-  }
-}
-
 static CURLcode
 schannel_connect_step2(struct Curl_cfilter *cf, struct Curl_easy *data)
 {
@@ -1303,18 +1264,28 @@ schannel_connect_step2(struct Curl_cfilter *cf, struct Curl_easy *data)
     Curl_safefree(inbuf[0].pvBuffer);
 
     /* check if the handshake was incomplete */
-    switch(sspi_status) {
-    case SEC_E_INCOMPLETE_MESSAGE:
+    if(sspi_status == SEC_E_INCOMPLETE_MESSAGE) {
       backend->encdata_is_incomplete = TRUE;
       connssl->io_need = CURL_SSL_IO_NEED_RECV;
       DEBUGF(infof(data,
                    "schannel: received incomplete message, need more data"));
       return CURLE_OK;
+    }
 
-    case SEC_I_CONTINUE_NEEDED:
-    case SEC_E_OK:
-      /* check if the handshake needs to be continued */
-      result = CURLE_OK;
+    /* If the server has requested a client certificate, attempt to continue
+       the handshake without one. This will allow connections to servers which
+       request a client certificate but do not require it. */
+    if(sspi_status == SEC_I_INCOMPLETE_CREDENTIALS &&
+       !(backend->req_flags & ISC_REQ_USE_SUPPLIED_CREDS)) {
+      backend->req_flags |= ISC_REQ_USE_SUPPLIED_CREDS;
+      connssl->io_need = CURL_SSL_IO_NEED_SEND;
+      DEBUGF(infof(data,
+                   "schannel: a client certificate has been requested"));
+      return CURLE_OK;
+    }
+
+    /* check if the handshake needs to be continued */
+    if(sspi_status == SEC_I_CONTINUE_NEEDED || sspi_status == SEC_E_OK) {
       for(i = 0; i < 3; i++) {
         /* search for handshake tokens that need to be send */
         if(outbuf[i].BufferType == SECBUFFER_TOKEN && outbuf[i].cbBuffer > 0) {
@@ -1328,35 +1299,47 @@ schannel_connect_step2(struct Curl_cfilter *cf, struct Curl_easy *data)
           if(result || (outbuf[i].cbBuffer != written)) {
             failf(data, "schannel: failed to send next handshake data: "
                   "sent %zu of %lu bytes", written, outbuf[i].cbBuffer);
-            result = CURLE_SSL_CONNECT_ERROR;
+            return CURLE_SSL_CONNECT_ERROR;
           }
         }
-      }
-      for(i = 0; i < 3; i++) {
+
         /* free obsolete buffer */
-        if(outbuf[i].pvBuffer)
+        if(outbuf[i].pvBuffer) {
           Curl_pSecFn->FreeContextBuffer(outbuf[i].pvBuffer);
+        }
       }
-      if(result)
-        return result;
-      break;
-
-    case SEC_I_INCOMPLETE_CREDENTIALS:
-      if(!(backend->req_flags & ISC_REQ_USE_SUPPLIED_CREDS)) {
-        /* If the server has requested a client certificate, attempt to
-           continue the handshake without one. This will allow connections to
-           servers which request a client certificate but do not require
-           it. */
-        backend->req_flags |= ISC_REQ_USE_SUPPLIED_CREDS;
-        connssl->io_need = CURL_SSL_IO_NEED_SEND;
-        DEBUGF(infof(data,
-                     "schannel: a client certificate has been requested"));
-        return CURLE_OK;
+    }
+    else {
+      char buffer[STRERROR_LEN];
+      switch(sspi_status) {
+      case SEC_E_INSUFFICIENT_MEMORY:
+        failf(data, "schannel: next InitializeSecurityContext failed: %s",
+              Curl_sspi_strerror(sspi_status, buffer, sizeof(buffer)));
+        return CURLE_OUT_OF_MEMORY;
+      case SEC_E_WRONG_PRINCIPAL:
+        failf(data, "schannel: SNI or certificate check failed: %s",
+              Curl_sspi_strerror(sspi_status, buffer, sizeof(buffer)));
+        return CURLE_PEER_FAILED_VERIFICATION;
+      case SEC_E_UNTRUSTED_ROOT:
+        failf(data, "schannel: %s",
+              Curl_sspi_strerror(sspi_status, buffer, sizeof(buffer)));
+        return CURLE_PEER_FAILED_VERIFICATION;
+#if 0
+      case SEC_E_INVALID_HANDLE:
+      case SEC_E_INVALID_TOKEN:
+      case SEC_E_LOGON_DENIED:
+      case SEC_E_TARGET_UNKNOWN:
+      case SEC_E_NO_AUTHENTICATING_AUTHORITY:
+      case SEC_E_INTERNAL_ERROR:
+      case SEC_E_NO_CREDENTIALS:
+      case SEC_E_UNSUPPORTED_FUNCTION:
+      case SEC_E_APPLICATION_PROTOCOL_MISMATCH:
+#endif
+      default:
+        failf(data, "schannel: next InitializeSecurityContext failed: %s",
+              Curl_sspi_strerror(sspi_status, buffer, sizeof(buffer)));
+        return CURLE_SSL_CONNECT_ERROR;
       }
-      FALLTHROUGH();
-
-    default:
-      return schannel_error(data, sspi_status);
     }
 
     /* check if there was additional remaining encrypted data */
@@ -1532,6 +1515,7 @@ schannel_connect_step3(struct Curl_cfilter *cf, struct Curl_easy *data)
   struct ssl_connect_data *connssl = cf->ctx;
   struct schannel_ssl_backend_data *backend =
     (struct schannel_ssl_backend_data *)connssl->backend;
+  struct ssl_config_data *ssl_config = Curl_ssl_cf_get_config(cf, data);
   CURLcode result = CURLE_OK;
   SECURITY_STATUS sspi_status = SEC_E_OK;
   CERT_CONTEXT *ccert_context = NULL;
@@ -1599,7 +1583,7 @@ schannel_connect_step3(struct Curl_cfilter *cf, struct Curl_easy *data)
 #endif
 
   /* save the current session data for possible reuse */
-  if(Curl_ssl_scache_use(cf, data)) {
+  if(ssl_config->primary.cache_session) {
     Curl_ssl_scache_lock(data);
     /* Up ref count since call takes ownership */
     backend->cred->refcount++;
@@ -1705,8 +1689,6 @@ enum schannel_renegotiate_caller_t {
   SCH_RENEG_CALLER_IS_SEND
 };
 
-#define MAX_RENEG_BLOCK_TIME (7 * 1000) /* 7 seconds in milliseconds */
-
 /* This function renegotiates the connection due to a server request received
    by schannel_recv. This function returns CURLE_AGAIN if the renegotiation is
    incomplete. In that case, we remain in the renegotiation (connecting) stage
@@ -1718,6 +1700,7 @@ schannel_recv_renegotiate(struct Curl_cfilter *cf, struct Curl_easy *data,
 {
   CURLcode result;
   curl_socket_t sockfd;
+  const timediff_t max_renegotiate_ms = 5 * 60 * 1000; /* 5 minutes */
   struct ssl_connect_data *connssl = cf->ctx;
   struct schannel_ssl_backend_data *backend =
     (struct schannel_ssl_backend_data *)connssl->backend;
@@ -1727,11 +1710,15 @@ schannel_recv_renegotiate(struct Curl_cfilter *cf, struct Curl_easy *data,
     failf(data, "schannel: unexpected call to schannel_recv_renegotiate");
     return CURLE_SSL_CONNECT_ERROR;
   }
-  DEBUGASSERT(caller <= SCH_RENEG_CALLER_IS_SEND);
+
   if(caller == SCH_RENEG_CALLER_IS_RECV)
     SCH_DEV(infof(data, "schannel: renegotiation caller is schannel_recv"));
-  else
+  else if(caller == SCH_RENEG_CALLER_IS_SEND)
     SCH_DEV(infof(data, "schannel: renegotiation caller is schannel_send"));
+  else {
+    failf(data, "schannel: unknown caller for schannel_recv_renegotiate");
+    return CURLE_SSL_CONNECT_ERROR;
+  }
 
   sockfd = Curl_conn_cf_get_socket(cf, data);
 
@@ -1758,7 +1745,7 @@ schannel_recv_renegotiate(struct Curl_cfilter *cf, struct Curl_easy *data,
     timediff_t elapsed;
 
     elapsed = curlx_timediff(curlx_now(), rs->start_time);
-    if(elapsed >= MAX_RENEG_BLOCK_TIME) {
+    if(elapsed >= max_renegotiate_ms) {
       failf(data, "schannel: renegotiation timeout");
       result = CURLE_SSL_CONNECT_ERROR;
       break;
@@ -1825,12 +1812,12 @@ schannel_recv_renegotiate(struct Curl_cfilter *cf, struct Curl_easy *data,
       }
 
       elapsed = curlx_timediff(curlx_now(), rs->start_time);
-      if(elapsed >= MAX_RENEG_BLOCK_TIME) {
+      if(elapsed >= max_renegotiate_ms) {
         failf(data, "schannel: renegotiation timeout");
         result = CURLE_SSL_CONNECT_ERROR;
         break;
       }
-      remaining = MAX_RENEG_BLOCK_TIME - elapsed;
+      remaining = max_renegotiate_ms - elapsed;
 
       if(blocking) {
         timeout = Curl_timeleft(data, NULL, FALSE);
@@ -2571,8 +2558,8 @@ static int schannel_init(void)
 #endif
   WINE_GET_VERSION_FN p_wine_get_version =
     CURLX_FUNCTION_CAST(WINE_GET_VERSION_FN,
-                        GetProcAddress(GetModuleHandleA("ntdll"),
-                                       "wine_get_version"));
+                        (GetProcAddress(GetModuleHandleA("ntdll"),
+                                        "wine_get_version")));
 #if defined(__clang__) && __clang_major__ >= 16
 #pragma clang diagnostic pop
 #endif
@@ -2598,7 +2585,7 @@ static void schannel_cleanup(void)
 
 static size_t schannel_version(char *buffer, size_t size)
 {
-  return curl_msnprintf(buffer, size, "Schannel");
+  return msnprintf(buffer, size, "Schannel");
 }
 
 static CURLcode schannel_random(struct Curl_easy *data,
